/* Driver for routine newt */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>

#define NRANSI
#define GAM 1.0e-5
#define NPTS 500  /*maximal number of points in the observed distribution*/
#define NN 100
#define NBMAX 1
#define MAXITS 200
#define TOLF 1.0e-4
#define TOLMIN 1.0e-6
#define TOLX 1.0e-7
#define STPMX 100.0
#define EPS 1.0e-4
#define TINY 1.0e-20;
#define ALF 1.0e-4
#define NR_END 1
#define FREE_ARG char*

static double dsqrarg;
#define DSQR(a) ((dsqrarg=(a)) == 0.0 ? 0.0 : dsqrarg*dsqrarg)

static double maxarg1, maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
        (maxarg1) : (maxarg2))


#define FREERETURN {free_dvector(fvec,1,n);free_dvector(xold,1,n);\
	free_dvector(p,1,n);free_dvector(g,1,n);free_dmatrix(fjac,1,n,1,n);\
	free_ivector(indx,1,n);return x;}

int N=2;
double *fvec;
int nbpoints;
double distr[NPTS], moy, prob;
int nn;

void (*nrfuncv)(int n, double v[], double f[]);

double gammaln(double a);
double psi(double a);
double comphgamma(double a);
double truncnegbin(double a, double b, int k);
double truncnegbinomial(double a, double b, int k);
double deriva(double a, double b, int k);
double derivb(double a, double b, int k);
void funcv(int n, double x[], double f[]);
double *newt( double x[], int n, int *check, void (*vecfunc)(int, double [], double []) );
void fdjac( int n, double x[], double **df, void (*vecfunc)(int, double [], double []) );
double fmin(double x[]);
void ludcmp(double **a, int n, int *indx, double *d);
void lubksb(double **a, int n, int *indx, double b[]);
void lnsrch(int n, double xold[], double fold, double g[], double p[], double x[], double *f, double stpmax, int *check, double (*func)(double []));
void nrerror(char error_text[]);
int *ivector(long nl, long nh);
double *dvector(long nl, long nh);
double **dmatrix(long nrl, long nrh, long ncl, long nch);
void free_ivector(int *v, long nl, long nh);
void free_dvector(double *v, long nl, long nh);
void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch);



/*function which computes the log of the gamma function*/
double gammaln(double a)
{
int j;
static int appel=0;
double x, y, yy, tmp, ser;
static double cof[6]={76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5};

	y=x=a;
	tmp=x+5.5;
	tmp-=( (x+0.5)*log(tmp) );
	ser=1.000000000190015;
	for(j=0; j<=5; j++) ser+=( cof[j]/++y );
	yy=( -tmp+log(2.5066282746310005*ser/x) );

return yy;
}

/*function which computes the value of the digamma function*/
double psi(double a)
{
double h, psi;

h=comphgamma(a);
psi=( gammaln(a+h)-gammaln(a) )/h;

/*printf("psi[%f]=%f,", a, psi);*/
return psi;
}

double comphgamma(double a)
{
double temp, h;

temp=a;
h=GAM*fabs(temp);
if(h==0.0) h=GAM;
a=temp+h;
h=a-temp;
a=temp;

return h;
}


/*function which computes the probabilities of the truncated negative binomial distribution*/
double truncnegbin(double a, double b, int k)
{
double G, R, V;

G=gammaln(a+k)/( gammaln(a)*gammaln(k+1) );
R=pow( (b/(1+b)), k );
V=pow( (1.0/(1+b)), a );

return G*R*V/(1-V);
}

/*function which computes the probability density of the truncated negative binomial distribution*/
double truncnegbinomial(double a, double b, int k)
{
static double R, q, w, truncnegbinomlaw;
double temp;
static int appel=0;


if(appel>=nbpoints) appel=0;

if(appel==0) {
	q=1+b;
	R=b/(1+b);
	w=1.0/( 1-pow(q, -a) );
	truncnegbinomlaw=w*pow(1/q, a);
	}
appel++;


temp=(a+k-1)*R*truncnegbinomlaw/k;
truncnegbinomlaw=temp;
return truncnegbinomlaw;
}


/*function which computes the derivative of the truncated negative binomial probability
with respect to a*/
double deriva(double a, double b, int k)
{
double x1, x2, x3, x4, temp2, temp3;

prob=truncnegbinomial(a, b, k);
/*printf("p[%d]=%.10f\n", k, prob);*/
temp2=log( 1.0/(1+b) );
temp3=pow( (1.0/(1+b)), a );

x1=psi(a+k)*prob;
x2=psi(a)*prob;
x3=temp2*prob;
x4=x3*temp3/(1-temp3);

return (x1-x2+x3+x4)/prob;
}

/*function which computes the derivative of the truncated negative binomial probability
with respect to b*/
double derivb(double a, double b, int k)
{
double x1, x2, x3, temp2, temp3, y;

/*printf("p[%d]=%.10f\n", k, prob);*/
temp2=1+b;
temp3=pow( (1.0/(1+b)), a );

y=(1.0/temp2) - (b/(temp2*temp2));
x1=prob*k*temp2*y/b;
x2=prob*a/temp2;
x3=x2*temp3/(1-temp3);

return (x1-x2-x3)/prob;
}


/*function which computes the derivatives of the likelihood with respect to a and b*/
void funcv(int n, double x[], double f[])
{
int i;

if(x[1]<=0.0) nrerror("Shape parameter alpha is negative or equal to 0.");
if(x[n]<=0.0) nrerror("Scale parameter beta is negative or equal to 0.");


f[1]=f[n]=0.0;
for(i=1; i<=nbpoints; i++) {
		f[1]+=distr[i]*deriva(x[1], x[n], i);
		f[n]+=distr[i]*derivb(x[1], x[n], i);
		/*printf("f1=%f f2=%f\n", f[1], f[n]);*/
			}
/*printf("f1=%f f2=%f\n", f[1], f[n]);*/
}


int main(void)
{
int check;
double *x;
FILE *infile, *outfile, *resultfile, *ajustfile;
char infilename[40], outfilename[40], resultfilename[40], ajustfilename[40];
int i, L;
double sumy, sumyx, sumyx2, invar, nbsites;
double freqinvar, mean, mean2, var, var2;
double Like, contrib;
double temp, temp2, temp3;

x=dvector(1, N);

printf("*************************************************************\n");
printf("<<Program AJUST: fits the observed distribution of changes.>>\n\n");


printf("Enter the initial values of the parameters a and b:\n");
for(i=1; i<=N; i++) {
		if(i==1) printf("a ? ");
		if(i==2) printf("b ? ");
		scanf("%lf", &x[i]);
		}


printf("What's the name of the distribution file ?\n");
scanf("%s", infilename);
L=strlen(infilename);
infilename[L]='\0';
if( (infile=fopen(infilename,"r"))==NULL ) {
		printf("can't open input file!\n");
		exit(-1);
		}

strncpy(ajustfilename, infilename, L-5);
ajustfilename[L-5]='\0';
strcat(ajustfilename, ".binom");
L=strlen(ajustfilename);
ajustfilename[L]='\0';
if( (ajustfile=fopen(ajustfilename,"w"))==NULL ) {
		printf("can't open output file for fitted distribution!\n");
		exit(-1);
		}
L=strlen(infilename);
strncpy(resultfilename, infilename, L-5);
resultfilename[L-5]='\0';
strcat(resultfilename, ".param");
L=strlen(resultfilename);
resultfilename[L]='\0';
if( (resultfile=fopen(resultfilename,"w"))==NULL ) {
		printf("can't open output file for displaying results!\n");
		exit(-1);
		}
strcpy(outfilename, "param");
L=strlen(outfilename);
outfilename[L]='\0';
if( (outfile=fopen(outfilename,"w"))==NULL ) {
		printf("can't open temporary output file!\n");
		exit(-1);
		}


i=0;
sumy=sumyx=sumyx2=0.0;
while( fscanf(infile, "%lf", &distr[i])!=EOF ) {
			sumy+=distr[i];
			sumyx+=(distr[i]*i);
			sumyx2+=(distr[i]*i*i);
			i++;
			if(i>NPTS) nrerror("NPTS exceeded in initialfunc.");
			}
invar=distr[0];
nbpoints=i-1;
nbsites=sumy;
freqinvar=invar/nbsites;
mean=sumyx/nbsites;
moy=sumyx/(nbsites-invar);
var=sumyx2/nbsites-mean*mean;
var2=sumyx2/(nbsites-invar)-moy*moy;

fprintf(resultfile, "observed distribution:\nmean=%f\nmean of the variable sites=%f\nvariance=%f\nvariance of the variable sites=%f\nnumber of invariant sites=%.1f\nproportion of invariant sites=%f\ntotal number of sites=%.1f\nmaximum number of changes=%d\n", mean, moy, var, var2, invar, freqinvar, nbsites, nbpoints);

x=newt(x, N, &check, funcv);
if(check) nrerror("Convergence problems.");

funcv(N, x, fvec);
fprintf(resultfile, "maximum-likelihood estimates of the model parameters:");
fprintf(resultfile, "\n%5s %7s %12s\n", "param", "solut", "1st deriv");
for(i=1; i<=N; i++) {
		if(i==1) fprintf(resultfile, "%3s  ", "a");
		if(i==2) fprintf(resultfile, "%3s  ", "b");
		fprintf(resultfile, "%9.6f  %10.6f\n", x[i], fvec[i]);
		}
fprintf(resultfile, "%3s  %9.6f  \n\n", "t", freqinvar);

fprintf(outfile, "%f\n%f\n%f\n%f\n", mean, x[1], x[2], freqinvar);

Like=distr[0]*log(freqinvar);
for(i=1;i<=nbpoints;i++) {
		contrib=log( (1-freqinvar)*truncnegbinomial(x[1],x[2],i) );
		/*printf("%f %f\n", distr[i]*contrib, contrib);*/
		Like+=distr[i]*contrib;
		}
fprintf(resultfile, "log-likelihood=%f\n", Like);

Like=0.0;
for(i=0;i<=nbpoints;i++) Like+=distr[i]*log(distr[i]/nbsites);
fprintf(resultfile, "max log-likelihood=%f\n", Like);

temp=1-pow(1+x[2],-x[1]);
mean=x[1]*x[2]/temp;
mean2=mean*mean;
fprintf(resultfile, "\nmean of the expected distribution for variable sites=%f\n", mean);
mean=(1-freqinvar)*mean;
fprintf(resultfile, "mean of the expected distribution for all sites=%f\n", mean);
temp2=x[1]*x[2]/(1+x[2]);
temp3=x[1]*x[2]*(1+x[2]);
var=( 1-temp2*(1.0/temp-1) )*temp3/temp;
fprintf(resultfile, "variance of the expected distribution for variable sites=%f\n", var);
var=(1-freqinvar)*var + freqinvar*(1-freqinvar)*mean2;
fprintf(resultfile, "variance of the expected distribution for all sites=%f\n", var);
fprintf(resultfile, "gamma distribution for all sites: a=%f, b=%f\n", mean*mean/var, mean/var);


fprintf(ajustfile, "%.1f\n", invar);
/*for(i=1;i<=nbpoints;i++) printf( "%f\n", truncnegbinomial(x[1],x[2],i) );*/
for(i=1;i<=nbpoints;i++) fprintf( ajustfile, "%.1f\n", nbsites*(1-freqinvar)*truncnegbinomial(x[1],x[2],i) );

free_dvector(x, 1, N);

fclose(infile);
fclose(outfile);
fclose(ajustfile);
fclose(resultfile);

return 0;
}


/*function which performs Newton's algorithm for solving nonlinear systems of equations*/
double *newt( double x[], int n, int *check, void (*vecfunc)(int, double [], double []) )
{
	void fdjac( int n, double x[], double **df, void (*vecfunc)(int, double [], double []) );
	double fmin(double x[]);
	void lnsrch( int n, double xold[], double fold, double g[], double p[], double x[], double *f, double stpmax, int *check, double (*func)(double []) );
	void lubksb(double **a, int n, int *indx, double b[]);
	void ludcmp(double **a, int n, int *indx, double *d);
	int i,its,j,*indx;
	double d,den,f,fold,stpmax,sum,temp,test,**fjac,*g,*p,*xold;

	indx=ivector(1,n);
	fjac=dmatrix(1,n,1,n);
	g=dvector(1,n);
	p=dvector(1,n);
	xold=dvector(1,n);
	fvec=dvector(1,n);
	nn=n;
	nrfuncv=vecfunc;
	f=fmin(x);
	test=0.0;     /*test for initial guess being a root*/
	for (i=1;i<=n;i++) if (fabs(fvec[i]) > test) test=fabs(fvec[i]);
	if (test<0.01*TOLF) FREERETURN
	for (sum=0.0,i=1;i<=n;i++) sum += DSQR(x[i]);
	stpmax=STPMX*FMAX(sqrt(sum),(double)n);
	for (its=1;its<=MAXITS;its++) {     /*START OF ITERATION LOOP FOR THE NEWTON'S ALGORITHM*/
		fdjac(n,x,fjac,vecfunc);
		for (i=1;i<=n;i++) {
			for (sum=0.0,j=1;j<=n;j++) {
					sum += fjac[j][i]*fvec[j];
					}
			g[i]=sum;
		}
		for (i=1;i<=n;i++) xold[i]=x[i];
		fold=f;
		for (i=1;i<=n;i++) p[i] = -fvec[i];
		ludcmp(fjac,n,indx,&d);    /*solve linear equations by LU decomposition*/
		lubksb(fjac,n,indx,p);
		lnsrch(n,xold,fold,g,p,x,&f,stpmax,check,fmin);  /*search of a suitable Newton step*/

		/*TESTS FOR CONVERGENCE*/
		test=0.0;      			for (i=1;i<=n;i++)
			if (fabs(fvec[i]) > test) test=fabs(fvec[i]);   /*convergence on function values (i.e. fvec sufficiently close to zero)*/
		if (test < TOLF) {
			*check=0;
			FREERETURN
		}
		if (*check) {
			test=0.0;
			den=FMAX(f,0.5*n);
			for (i=1;i<=n;i++) {
				temp=fabs(g[i])*FMAX(fabs(x[i]),1.0)/den;
				if (temp > test) test=temp;
			}
			*check=(test < TOLMIN ? 1 : 0);
			FREERETURN
		}
		test=0.0;
		for (i=1;i<=n;i++) {    /*convergence on Newton step*/
			temp=(fabs(x[i]-xold[i]))/FMAX(fabs(x[i]),1.0);
			if (temp > test) test=temp;
		}
		if (test < TOLX) FREERETURN
		/*printf(".");*/
	}
	nrerror("MAXITS exceeded in newt.");
	return x;
}


/*function which calculates the Jacobian matrix of derivatives*/
void fdjac( int n, double x[], double **df, void (*vecfunc)(int, double [], double []) )
{
	int i,j;
	double *f, temp, hh;

	f=dvector(1,n);

for(j=1; j<=n; j++) {
	temp=x[j];
	hh=EPS*fabs(temp);
	if (hh == 0.0) hh=EPS;
	x[j]=temp+hh;
	hh=x[j]-temp;
	(*vecfunc)(n, x, f);
	x[j]=temp;
	for(i=1; i<=n; i++) df[i][j]=(f[i]-fvec[i])/hh;
	}

	free_dvector(f,1,n);
}


/*function which computes f=1/2*F*F, where F is the Jacobian matrix of derivatives which was computed by the function fdjac. During the solving process, one must minimize f. This is done by the function lnsrch*/
double fmin(double x[])
{
	int i;
	double sum;


	(*nrfuncv)(nn,x,fvec);
	for (sum=0.0,i=1;i<=nn;i++) sum += DSQR(fvec[i]);
	return 0.5*sum;
}


/*function which performs the LU decomposition of the input matrix a. This is the first step for solving linear systems of equations by the LU method*/
void ludcmp(double **a, int n, int *indx, double *d)
{
	int i,imax,j,k;
	double big,dum,sum,temp;
	double *vv;

	vv=dvector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine ludcmp.");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {    /*loop of Crout's algorithm for decomposing a matrix*/
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free_dvector(vv,1,n);
}


/*function which performs forward and backward substitution. This is a method for solving systems of equations. The input matrix is a matrix decomposed with the LU method performed by the function ludcmp*/
void lubksb(double **a, int n, int *indx, double b[])
{
	int i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {    /*forward substitution*/
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {    /*backsubstitution*/
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}


/*function which performs "line search and backtracking". This is the strategy for minimizing the f value computed by the function fmin*/
void lnsrch(int n, double xold[], double fold, double g[], double p[], double x[], double *f, double stpmax, int *check, double (*func)(double []))
{
	int i;
	double a,alam,alam2,alamin,b,disc,f2,fold2,rhs1,rhs2,slope,sum,temp,
		test,tmplam;

	*check=0;
	for (sum=0.0,i=1;i<=n;i++) sum += p[i]*p[i];
	sum=sqrt(sum);
	if (sum > stpmax)
		for (i=1;i<=n;i++) p[i] *= stpmax/sum;
	for (slope=0.0,i=1;i<=n;i++)
		slope += g[i]*p[i];
	test=0.0;
	for (i=1;i<=n;i++) {
		temp=fabs(p[i])/FMAX(fabs(xold[i]),1.0);
		if (temp > test) test=temp;
	}
	
	alamin=TOLX/test;
	alam=1.0;       /*always try full Newton step first*/
	alam2=alam;
	f2=*f;
	fold2=fold;
	for (;;) {
		for(i=1; i<=n; i++) {
			x[i]=xold[i]+alam*p[i];
			/*printf("x=%f", x[i]);*/
			}
		*f=(*func)(x);
		if (alam < alamin) {
			for (i=1;i<=n;i++) x[i]=xold[i];
			*check=1;
			return;
				} else if (*f <= fold+ALF*alam*slope) return;  /*sufficient decrease in function f (i.e. fmin)*/
		else {         /*BACKTRACK*/
			if (alam == 1.0)
				tmplam = -slope/(2.0*(*f-fold-slope));
			else {
				rhs1 = *f-fold-alam*slope;
				rhs2=f2-fold2-alam2*slope;
				a=(rhs1/(alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
				b=(-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
				if (a == 0.0) tmplam = -slope/(2.0*b);
				else {
					disc=b*b-3.0*a*slope;
					if (disc<0.0) nrerror("Roundoff problem in lnsrch.");
					else tmplam=(-b+sqrt(disc))/(3.0*a);
				}
				if (tmplam>0.5*alam)
					tmplam=0.5*alam;
			}
		}
		alam2=alam;
		f2 = *f;
		fold2=fold;
		alam=FMAX(tmplam,0.1*alam);
	}
}



/*Numerical Recipes Tools (extracted from utility file nrutil.c)*/

void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"Try to input other initial values.\n");
	fprintf(stderr,"...now exiting to system...\n");
	exit(1);
}

int *ivector(long nl, long nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
	int *v;

	v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
	if (!v) nrerror("allocation failure in ivector()");
	return v-nl+NR_END;
}

double *dvector(long nl, long nh)
/* allocate a double vector with subscript range v[nl..nh] */
{
	double *v;

	v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
	if (!v) nrerror("allocation failure in dvector()");
	return v-nl+NR_END;
}

double **dmatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */
{
	long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
	double **m;

	/* allocate pointers to rows */
	m=(double **) malloc((size_t)((nrow+NR_END)*sizeof(double*)));
	if (!m) nrerror("allocation failure 1 in matrix()");
	m += NR_END;
	m -= nrl;

	/* allocate rows and set pointers to them */
	m[nrl]=(double *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
	if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
	m[nrl] += NR_END;
	m[nrl] -= ncl;

	for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

	/* return pointer to array of pointers to rows */
	return m;
}

void free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */
{
nh=0L;
	free((FREE_ARG) (v+nl-NR_END));
}

void free_dvector(double *v, long nl, long nh)
/* free a double vector allocated with dvector() */
{
nh=0L;
	free((FREE_ARG) (v+nl-NR_END));
}

void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch)
/* free a double matrix allocated by dmatrix() */
{
nrh=nch=0L;
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}


#undef MAXITS
#undef TOLF
#undef TOLMIN
#undef STPMX
#undef FREERETURN
#undef EPS
#undef TINY
#undef ALF
#undef TOLX
#undef NRANSI
