#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAX 1000
#define nmlngth 10   /*number of characters in a species name*/

int translate(char, char, char);


/*function which reads the name of one sequence only in new phylip format*/
char *newname(FILE *infile)
{
char *nom;
int i, j, c;

/*allocate memory for the pointer nom */
if( (nom=(char*)malloc( (nmlngth+1)*sizeof(char)) )==NULL ) {
				printf("not enough memory\n");
				exit(1);
				}

/*read the name of the sequence*/
i=j=0;
while(i<nmlngth) {
	c=fgetc(infile);
	if(c=='\n') continue;
	if(c==' ') {
		i++;
		continue;
		}
	nom[i]=c;
	i++;
	j++;
	}
nom[j]='\0';

return nom;
}

/*function which reads all the sequences in new phylip format*/
void newseq(FILE *infile, int Nseq, int Nsites, char **seq, char **seqname, int Opt)
{
int i, j, k, x;
int c;

/*allocate memory for all the pointers seq[i] */
for(i=0; i<Nseq; i++) {
			seq[i]=(char *)malloc((Nsites+1)*sizeof(char));
			if(seq[i]==NULL) {
					printf("not enough memory\n");
					exit(1);
					}
			}

/*read all the sequences in new format*/
i=j=x=0;
while(i<Nseq) {
		if(x==0) seqname[i]=newname(infile);
		k=0;
		while( (c=fgetc(infile))!='\n' ) {
					if(c==' ') continue;
					if(islower(c)!=0) c=toupper(c);
					if(c=='U') c='T';
					seq[i][j]=c;
					k++;
					j++;
					}
		if(k==0) continue;
		if(i<(Nseq-1)) j=j-k;
		if(i==(Nseq-1)) {
			if(j!=Nsites) i=(-1);
			x++;
			}
		i++;
		}

/*put \0 at the end of all the sequences*/
for(i=0; i<Nseq; i++) seq[i][j]='\0';
}


/*function which reads  the name of one sequence only in old phylip format*/
char *oldname(FILE *infile)
{
char *nom;
int i, j, c;

/*allocate memory for the pointer nom */
if( (nom=(char*)malloc( (nmlngth+1)*sizeof(char)) )==NULL) {
				printf("not enough memory\n");
				exit(1);
				}

/*read the name of the sequence*/
i=j=0;
while(i<nmlngth) {
	c=fgetc(infile);
	if(c=='\n') break;
	if(c==' ') {
		i++;
		continue;
		}
	nom[i]=c;
	i++;
	j++;
	}
nom[j]='\0';

return nom;
}


/*function which reads one sequence only in old phylip format*/
char *oldseq(FILE *infile, int Nsites, int Opt)
{
char *seq;
char line[MAX+1];
int i;
int c;

/*allocate memory for the pointer sequence*/
if( (seq=(char*)malloc((Nsites+1)*sizeof(char)) )==NULL) {
				printf("not enough memory\n");
				exit(1);
				}

/*read the sequence*/
i=0;
while(i<Nsites)	{
		c=fgetc(infile);
		if(c=='\n' || c==' ') continue;
		if(islower(c)!=0) c=toupper(c);
		if(c=='U') c='T';	  
		seq[i]=c;
		i++;
		}

/*put \0 at the end of the sequence*/
seq[i]='\0';

/*read the end of the sequence line*/
if( (fgets(line, MAX, infile))==NULL ) {
					printf("err_fgets\n");
					exit(1);
					}

return seq;
}


/*don't take into account undefined nucleotides*/
int epur(int Nseq, int Nsites, char **seq, char **seqname, int typedata, int clear) {
int i, j, jj, biz, aa;
char c, c1, c2, c3;

if(typedata==2) {
	for(i=0; i<Nseq; i++) for(j=0; j<Nsites; j++) {
				c=seq[i][j];
				if(c!='A' && c!='C' && c!='G' && c!='T') {
					if(c!='-') {
						printf("undefined nucleotide at position %d of the sequence %s\n", j+1, seqname[i]);
						if(clear==1) seq[i][j]='-';
						}
					}
				}
	}

if(typedata==1) {
	for(i=0; i<Nseq; i++) for(j=2; j<Nsites; j+=3) {
		biz=0;
		c1=seq[i][j-2]; c2=seq[i][j-1]; c3=seq[i][j];
		if(c1!='A' && c1!='C' && c1!='G' && c1!='T') {
			if(c1!='-') printf("undefined nucleotide at position %d of the sequence %s\n", j-2+1, seqname[i]);
			biz=1;
			}
		if(c2!='A' && c2!='C' && c2!='G' && c2!='T') {
			if(c2!='-') printf("undefined nucleotide at position %d of the sequence %s\n", j-1+1, seqname[i]);
			biz=1;
			}
		if(c3!='A' && c3!='C' && c3!='G' && c3!='T') {
			if(c3!='-') printf("undefined nucleotide at position %d of the sequence %s\n", j+1, seqname[i]);
			biz=1;
			}
		if(biz==0) {
			aa=translate(seq[i][j-2],seq[i][j-1],seq[i][j]);
			if(aa==21) {
				printf("stop codon in positions %d,%d,%d of the sequence %s\n", j-2+1, j-1+1, j+1, seqname[i]);
				biz=1;
				}
			}
		if(biz==1) seq[i][j-2]=seq[i][j-1]=seq[i][j]='-';
		}
	}


/*eliminate all undefined positions*/
biz=0;
for(j=0; j<Nsites; j++) for(i=0; i<Nseq; i++) {
				if(seq[i][j]=='-') {
					biz=1;
					if(j!=Nsites-1) for(i=0; i<Nseq; i++) for(jj=j; jj<Nsites-1; jj++) seq[i][jj]=seq[i][jj+1];
					Nsites-=1;
					j=j-1;
					break;
					}
				}
for(i=0; i<Nseq; i++) seq[i][Nsites]='\0';

if(typedata==2 && biz==1) {
	printf("\nAfter removal of gaps and/or undefined nucleotides,\n");
	printf("the actual number of sites used is %d\n\n", Nsites);
	}
if(typedata==1 && biz==1) {
	printf("\nGaps and/or stop codons and/or undefined nucleotides were found and then eliminated in all sequences.\n");
	printf("Thus, the actual number of sites used is %d\n\n", Nsites);
	}

return Nsites;
}
