/*Studier-Keppler*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX 1000
#define NSM 4000  /*maximum number of sequences allowed*/

int nsm;
int n; double *d[NSM]; char *nm[NSM];
int ns, nf, *id, *t; double *la;
int *droit, *gauche;
int root=0;
FILE *outfile;
FILE *infile;


void Rij(double **d, int n);
void Sij(int *ii, int *jj);
void reduction(int ii, int jj);
void sommets();
void LirTab(FILE *infile, double **d, char **nm, int *n);
void SymTab(double **d, int n);
void cherchedesc(int *t, int ns, int node, int *gauche, int *droit);
void printparent(int node, int *gauche, int *droit, int nf, char **nm, double *la);


void Rij(double **d, int n)
{
/*place the sum of the distances between i and all others in d[i][i].
*d[] must be symmetrized before*/

  int i,j; double smd;

  smd=0;
  for(j=1;j<=n;j++){ 
    		smd=0;
    		for(i=1;i<=n;i++) if(i!=j) smd+=d[i][j];
    		d[j][j]=smd;
  		}     
}

void Sij(int *ii, int *jj)
{
  int i,j; double sij,smin;
 
  smin=0;

  for(i=1;i<=n;i++)
    for(j=i+1;j<=n;j++){
  
      	sij=d[i][j]*(n-2)-d[i][i]-d[j][j];

	if(sij<=smin) {*ii=i; *jj=j; smin=sij;}   /*search for the smallest sij*/
    	}
/*printf("smin=%f\n", smin);*/
  
  ns+=1; t[id[*ii]]=ns; t[id[*jj]]=ns;

 
  la[id[*ii]]=((n-2)*d[*ii][*jj]+d[*ii][*ii]-d[*jj][*jj])/(2*(n-2));
  la[id[*jj]]=((n-2)*d[*ii][*jj]-d[*ii][*ii]+d[*jj][*jj])/(2*(n-2));
/*printf("%f %f, ", la[id[*ii]],la[id[*jj]]);*/
}

void reduction(int ii,int jj)
{
  int i;

  for(i=1;i<=n;i++) if(i!=ii && i!=jj){
     			d[i][ii]=(d[i][ii]+d[i][jj]-d[ii][jj])/2; d[ii][i]=d[i][ii];
   			 }
  
  d[ii][ii]=0; id[ii]=ns; 

  for(i=1;i<n;i++){				/* n->jj */
    	d[i][jj]=d[i][n]; d[jj][i]=d[n][i];
  	}

  d[jj][jj]=0; id[jj]=id[n];
       
  n-=1;  
}

void sommets(void)
{
  int i,j,k;

  i=id[1]; j=id[2]; k=id[3];
  ns+=1; t[i]=t[j]=t[k]=ns;

  
  la[i]=(d[1][2]+d[1][3]-d[2][3])/2;
  la[j]=(d[1][2]+d[2][3]-d[1][3])/2;
  la[k]=(d[1][3]+d[2][3]-d[1][2])/2;
/*printf("som=%f %f %f, ", la[i], la[j], la[k]);*/
}

void LirTab(FILE *infile, double **d, char **nm, int *n)
{

   /*d[i][j] i->columns j->lines*/


  int kk; int i,j; char nom[21];
  char line[MAX+1];


  fgetc(infile);

  for(i=1;i<=nsm-1;i++){
		j=i+1;
		while(j<=nsm) {	     /*read the distances*/
				fscanf(infile, "%lf", &d[i][j]);
				/*printf("%f ", d[i][j]);*/
				j++;
				}
		}


  fgets(line,MAX,infile);
  while( (kk=fgetc(infile))=='\n' ) fgetc(infile);
  fgets(line,MAX,infile);

  for(i=1; i<=nsm; i++) {
		fscanf(infile, "%20s", nom);                 /*read the species names*/
		/*printf("nom=%s\n", nom);*/
		if( (nm[i]=(char *)malloc( (strlen(nom)+1)*sizeof(char) ))==NULL ) {
						printf("not enough memory for array nm\n");
						exit(-2);
						}
		strcpy(nm[i],nom);
		}

/*for(i=1;i<=nsm;i++) printf("%d %s\n", i, nm[i]);*/

  fgetc(infile);

  *n=nsm;
/*printf("n=%d\n", n);*/

}

void SymTab(double **d, int n)
{
/*copy lower triangle into upper triangle (i.e., symmetrize the distance matrix)*/

  int i,j;

  for(j=1;j<=n;j++) for(i=1;i<j;i++) d[j][i]=d[i][j]; 
}


int main(int argc, char *argv[])
{
  int it,i,ii,jj,imp;
  char fil[50], outfil[50];
  int compt, node1, node2, node3;
  int L;
  double sumla;
  
  printf("****************************************************************\n");
  printf("<<Program SK: applies Studier/Keppler tree-building algorithm.>>\n");

  /*printf("\nprintout (type <1> to write the description of the clustering cycles in the tree file, else type <0>): ");
  scanf("%d", &imp);*/
	imp=0;

 /*if(argc>1)*/ strcpy(fil, argv[1]);
 /*printf("\ninput distance file: ");
 scanf("%s", fil);*/
 L=strlen(fil);
 fil[L]='\0';
 if( (infile=fopen(fil,"r"))==NULL ) {
	fprintf(stderr, "cannot open input file!\n");
	exit(-1);
	}

  fscanf(infile, "%d", &nsm);              /*read the number of sequences*/
  /*printf("nsm=%d\n", nsm);*/


/*memory allocations for arrays*/
/*for array *id, subscript numbering goes from 1 to nsm, not from 0 to nsm-1. Thus, nsm+1 elements are allocated in order to compute id[nsm]*/
if( (id=(int *)malloc( (nsm+1)*sizeof(int) ))==NULL ) {
					printf("not enough memory for array id\n");
					exit(-2);
					}
if( (t=(int *)malloc( (2*nsm)*sizeof(int) ))==NULL ) {
					printf("not enough memory for array t\n");
					exit(-2);
					}
if( (la=(double *)malloc( (2*nsm)*sizeof(double) ))==NULL ) {
					printf("not enough memory for array la\n");
					exit(-2);
					}

/*WARNING: arrays *gauche and *droit are made up of nsm elements, but subscript numbering goes from nf to ns. This explains why we use gauche[i-nf] and droit[i-nf] below*/
if( (droit=(int *)malloc( nsm*sizeof(int) ))==NULL ) {
					printf("not enough memory for array droit\n");
					exit(-2);
					}
if( (gauche=(int *)malloc( nsm*sizeof(int) ))==NULL ) {
					printf("not enough memory for array gauche\n");
					exit(-2);
					}
for(i=1;i<=nsm;i++) {
		  /*for each array d[i], the same strategy as for the array *id above is used*/
		if( (d[i]=(double *)malloc( (nsm+1)*sizeof(double) ))==NULL ) {
					printf("not enough memory for array d\n");
					exit(-2);
					}
		}

		
LirTab(infile,d,nm,&n);
/*printf("n=%d\n", n);*/

  ns=nf=n; for(i=1;i<=n;i++) id[i]=i;

  it=0;
  while (n>3){
    		it+=1;
    		SymTab(d,n); 
		Rij(d,n);
		/*printf("ii=%d jj=%d\n", ii, jj);*/
    		Sij(&ii,&jj);
    		reduction(ii,jj);  
  		}
  sommets();

 
  for(i=1;i<=ns;i++) if(la[i]<=0.0) la[i]=0.0;  /*NO NEGATIVE BRANCH LENGTHS!!*/

  
  /*printf("\nenter a name for the tree file: ");*/
  strncpy(outfil, fil, L);
  outfil[L]='\0';
  strcat(outfil, "sk");
  L=strlen(outfil);
  outfil[L]='\0';
  if( (outfile=fopen(outfil,"w"))==NULL ) {
	fprintf(stderr, "cannot open output file!\n");
	exit(-1);
	}

for(i=nf+1; i<ns; i++) cherchedesc(t, ns, i, gauche, droit);

for(compt=0,i=1; i<ns; i++) if(t[i]==ns) {
				if(compt==0) {
					node1=i;
					compt=1;
					}
				else if(compt==1) {
					node2=i;
					compt=2;
					}
				else if(compt==2) {
					node3=i;
					compt=3;
					}
				else break;
				}

fprintf(outfile, "(");
fprintf(outfile, "(");
if(node1<=nf) {
	fprintf(outfile, "%s:", nm[node1]);
	fprintf(outfile, "%.6f", la[node1]);
	}
else printparent(node1, gauche, droit, nf, nm, la);
fprintf(outfile, ",");
if(node2<=nf) {
	fprintf(outfile, "%s:", nm[node2]);
	fprintf(outfile, "%.6f", la[node2]);
	}
else printparent(node2, gauche, droit, nf, nm, la);
fprintf(outfile, ")");
fprintf(outfile, ":%.6f", 0.5*la[node3]);
root=1;
fprintf(outfile, ",");
if(node3<=nf) {
	fprintf(outfile, "%s:", nm[node3]);
	fprintf(outfile, "%.6f", 0.5*la[node3]);
	}
else printparent(node3, gauche, droit, nf, nm, la);
fprintf(outfile, ");");


if(imp==1) {
	for(i=nf+1; i<ns; i++) {
			fprintf(outfile, "\n\nCycle %4d   ", i-nf);
			if(gauche[i-nf]<=nf) fprintf(outfile, " OTU %4d (%6.4f)", gauche[i-nf], la[gauche[i-nf]]);
			else fprintf(outfile, "node %4d (%6.4f)", gauche[i-nf], la[gauche[i-nf]]);
			fprintf(outfile, "  and  ");
			if(droit[i-nf]<=nf) fprintf(outfile, " OTU %4d (%6.4f)", droit[i-nf], la[droit[i-nf]]);
			else fprintf(outfile, "node %4d (%6.4f)", droit[i-nf], la[droit[i-nf]]);
			fprintf(outfile, "  grouped in %4d\n", i);
			if(gauche[i-nf]<=nf) fprintf(outfile, "             %-25s", nm[gauche[i-nf]]);
			if(droit[i-nf]<=nf) fprintf(outfile, "%s", nm[droit[i-nf]]);
			}
	fprintf(outfile, "\n\nLast cycle:\n\n");
	if(node1<=nf) fprintf(outfile, " OTU %4d (%6.4f)", node1, la[node1]);
	else fprintf(outfile, "node %4d (%6.4f)", node1, la[node1]);
	fprintf(outfile, " , ");
	if(node2<=nf) fprintf(outfile, " OTU %4d (%6.4f)", node2, la[node2]);
	else fprintf(outfile, "node %4d (%6.4f)", node2, la[node2]);
	fprintf(outfile, "  and  ");
	if(node3<=nf) fprintf(outfile, " OTU %4d (%6.4f)", node3, la[node3]);
	else fprintf(outfile, "node %4d (%6.4f)", node3, la[node3]);
	fprintf(outfile, "  grouped in %4d\n", ns);
	if(node1<=nf) fprintf(outfile, "%-22s", nm[node1]);
	if(node2<=nf) fprintf(outfile, "%-25s", nm[node2]);
	if(node3<=nf) fprintf(outfile, "%s", nm[node3]);
	}

  printf("\n");

sumla=0.0;
for(i=1;i<=ns;i++) {
		/*if(i<=nf) printf("id=%d, ", id[i]);*/
		if(i<ns) {
			/*printf("la=%f, ", la[i]);*/
			sumla+=la[i];
			}
		/*if(i>nf && i<ns) printf("t=%d, ", t[i]);
		if(i>nf && i<ns) printf("droit(%d)=%d, ", i, droit[i-nf]);
		if(i>nf && i<ns) printf("gauche(%d)=%d, ", i, gauche[i-nf]);
		if(i<ns) printf("t=%d, ", t[i]);*/
		}
printf("total length of the tree (%d branches)=%f\n", 2*nsm-3, sumla);


printf("tree file created: %s\n", outfil);
fclose(infile);
fclose(outfile);

return 0;
}


/*function which searches for the 2 descendants of each internal node*/
void cherchedesc(int *t, int ns, int node, int *gauche, int *droit)
{
int i;
int compt=0;

for(i=1; i<ns; i++) if(t[i]==node) {
				if(compt==0) {
					gauche[node-nf]=i;
					compt=1;
					}
				else if(compt==1) {
					droit[node-nf]=i;
					compt=2;
					}
				else break;
				}
}

/*function which writes the tree in the parenthesized notation*/
void printparent(int node, int *gauche, int *droit, int nf, char **nm, double *la)
{
static int ouv=0, fer=0;

if(node>nf) {
	fprintf(outfile, "(");
	ouv++;
	if(gauche[node-nf]<=nf) {
		fprintf(outfile, "%s:", nm[gauche[node-nf]]);
		fprintf(outfile, "%.6f", la[gauche[node-nf]]);
		}
	else printparent(gauche[node-nf], gauche, droit, nf, nm, la);
	fprintf(outfile, ",");
	if(droit[node-nf]<=nf) {
			fprintf(outfile, "%s:", nm[droit[node-nf]]);
			fprintf(outfile, "%.6f", la[droit[node-nf]]);
			}
	printparent(droit[node-nf], gauche, droit, nf, nm, la);
	fprintf(outfile, ")");
	fer++;
	if(root==1 && ouv==fer) fprintf(outfile, ":%.6f", 0.5*la[node]);
	else fprintf(outfile, ":%.6f", la[node]);
	}

}

